# File: eventchannel.py
# Author: Kapil Thangavelu <kvthan@wm.edu>
# License: GPL <www.fsf.org>
# CVS: $Id: eventchannel.py,v 1.3 2001/10/02 18:48:06 ender Exp $
# 

from OFS.SimpleItem import Item
from Globals import DTMLFile, InitializeClass
from AccessControl import ClassSecurityInfo
from App.Extensions import getObject
from Shared.DC.ZRDB.TM import TM
from Acquisition import Implicit

from eventchannel_exceptions import *


if 1:
    class Log:
        def debug(self,msg):pass
        def warning(self,msg):pass
        def critical(self,msg):pass
    log = Log()
    del Log
else:    
    from eventlog import log            

from ZODB.PersistentMapping import PersistentMapping

from BTrees import IOBTree
from BTrees import OIBTree




#log = None
TRANSACTION_PUBLISH_LIMIT = 5
        
NOTES="""
Need to be able to go from
uid -> channel
channel -> uid
channel, uid -> filter

one weakness is removing event_types
won't remove registrations although
it will stop publishing.

to fix this prune_registrations
will do the required bookeeping,
why not include it, because its
an iterative journey through
all data structures maintained
by the channel.

filter checks, allow filters
to be python functions
from product modules,
... checkout external methods
for allowing it from extensions
subdir

event obj?
  should i wrap event sources
  to prevent assignment?
  (since assignment would be
  to a aq wrapper most likely
  this could be a leak
  source )

thinking about application integration
need to do some work for remapping
channels.

probably should be more forgiving
about duplicates registrations 
and silently fail...?

migration to Btrees 

propertymanager with configurable
transaction broadcasts and multiple
object registrations for eventtype with
different listener.
"""

manage_addEventChannelForm = DTMLFile('dtml/manage_addEventChannelForm',
                                      globals())

def manage_addEventChannel(self, id, title, REQUEST=None):
    """ oh illustrious publishing string""" 

    
    ob = EventChannel(id, title)
    self._setObject(str(id), ob)

    if REQUEST:
        self.manage_main(self, REQUEST, update_menu=1)


def get_uid(obj):
    if type(obj) == type(''): 
        # should probably verify that infact the object exists
        # however string types uids should generally be passed 
        # in only by internal methods
        return obj
    try: uid = obj.getPhysicalPath
    except AttributeError:
        raise TypeError(
            "An event channel object must support the 'getPhysicalPath' "
            "method if no unique id is provided when cataloging"
            )
    return '/'.join(uid())

#log.debug('loading')

class EventChannel(Item,TM, Implicit):

    security = ClassSecurityInfo()
    
    meta_type = 'Event Channel'
    
    manage_options = (
                      {'label':'Event Types', 'action':'manage_eventTypes',
                      },
                      {'label':'Listeners', 'action':'manage_eventListeners',
                      },
                      {'label':'Sources', 'action':'manage_eventSources',
                      }
                      )
    
    security.declareProtected('Manage Event Channel', 'manage_eventTypes')
    manage_eventTypes = DTMLFile('dtml/manage_eventTypes', globals())
    
    security.declareProtected('Manage Event Channel', 'manage_eventListeners')    
    manage_eventListeners = DTMLFile('dtml/manage_eventListeners', globals())
    
    security.declareProtected('Manage Event Channel', 'manage_eventSources')        
    manage_eventSources = DTMLFile('dtml/manage_eventSources', globals())
    
    def __init__(self, id, title):
        
        self.id = str(id)
        self.title = str(title)
        
        # apparently more magic incantations are involved than
        # simply p_changed = 1. may the gods of persistence
        # smile upon your efforts such that you may drink
        # of the river lethe and forget the tale of ozymandias.
        self._event_l_methods = PersistentMapping()
        self._event_types     = PersistentMapping()
        self._event_listeners = PersistentMapping()
        self._event_filters   = PersistentMapping()
        self._event_sources   = PersistentMapping()
        
        log.debug("creating channel")

    security.declareProtected('Use EventChannel', 'register_listener') 
    def register_listener(self, object, method, event_type, filter=None):
        """ register an object for an event channel with an optional
        filter. """
        
        uid = get_uid(object)

        # event type valid?
        try: listeners = self._event_types[event_type]
        except KeyError:
            raise BadEventType (
                "This event type does not exist %s"%event_type
                )
        # method valid?
        try:
            methodobj = getattr(object.aq_base, method)
            assert hasattr(methodobj, 'func_code')==1
        except Exception, e:
            #print exception in method', Exception, e
            raise AttributeError (
                " event listener does not have %s method" %method
                )

        # are there cases where we want to have one object listen
        # to the same event with multiple targets??
        # should probably make this configurable

        # an object can only register once for an event_type...
        # cannot register multiple filters to be called by the same object
        # this could be a problem with subclassing.
        # should reconsider after more real world experience.

        if listeners.count(uid) > 0:
            raise BadEventListener ("listener already registered for this event")
        
        # filter valid? raises error if not
        # else returns 1
        if filter:

            valid = self._validate_filter(object, filter)
            
            # add listener, type -> filter assoc
            self._event_filters[(uid, event_type)] = filter            
            
        # add event_type -> listener assoc
        listeners.append(uid)
        
        # trigger p machinery
        self._event_types[event_type]=listeners
        
        # add listener -> event_type assoc
        registered_et = self._event_listeners.get(uid, None)
        if registered_et:
            registered_et.append(event_type)
        else:
            self._event_listeners[uid]=[event_type]

        # add listener, type -> method assoc
        self._event_l_methods[(uid, event_type)] = method
        
        self._p_changed = 1
        
    security.declareProtected('Use EventChannel', 'unregister_listener')
    def unregister_listener(self, object, event_type=None):
        """ unregister a listener on an event channel(s),
        unless a channel is specified all registrations are
        removed """
        
        uid=get_uid(obj)

        if not self._event_listeners.has_key(uid):
            return
        
        # remove from et listener q
        if event_type:
            if not self._event_listeners[uid].count(event_type) > 0:
                return 
            # p machinery ...
            listener_et = self._event_listeners[uid]
            listener_et.remove(event_type)
            
            if len(listener_et) == 0:
                del self._event_listeners[uid]
            else:
                self._event_listeners[uid]=listener_et
                
            event_types = [event_type]
        else:
            event_types = self._event_listeners[uid]
            del self._event_listeners[uid]

        # remove methods and filters
        for et in event_types:
            if self._event_filters.has_key((uid, et)):
                del self._event_filters[(uid, et)]                
            if self._event_l_methods.has_key( (uid, et) ):
                del self._event_l_methods[ (uid, et) ]
            else:
                log.warning('removed listener with no target method')
                
        self._p_changed = 1



    security.declareProtected('Use EventChannel', 'set_listener_filter_for_event_type')
    def set_listener_filter_for_event_type(self, object, filter, event_type):
        " "
        if not self._event_types.has_key(event_type):
            return
        if type(object) == type(''):
            object = self.unrestrictedTraverse(object)
        self.validate_filter(object, filter)
        self._event_filters[ (get_uid(object), event_type)] = filter 
        
    security.declarePublic('is_event_listener')
    def is_event_listener(self, object, event_type=None):
        """ should we allow the same object to register
        more than once for an event type with different
        methods.... no"""
        sid = get_uid(object)
        if event_type is None:
            return self._event_listeners.has_key(sid)
        elif self._event_listeners.has_key(sid) and \
             self._event_types.has_key(event_type):
            return self._event_types[event_type].count(sid) 
        return 0
    
    security.declareProtected('Use EventChannel','get_event_types_for_listener')
    def get_event_types_for_listener(self, object):
        uid = get_uid(object)
        # have to be careful about returning mutable types
        # that might change registrations in event channel

        #print 'getting event types for %s'%uid
        r = self._event_listeners.get(uid, ())
        #print r
        return tuple(self._event_listeners.get(uid, ()))
    
    security.declareProtected('Use EventChannel', 'get_listener_registration')
    def get_listener_registration(self, object):
        uid = get_uid(object)
        pass
    
    security.declarePublic('get_event_listeners')
    def get_event_listeners(self):
        return self._event_listeners.keys()
        
    security.declareProtected('Use EventChannel', 'register_source')
    def register_source(self, object, event_type):
        """ add a source for a given event channel """

        uid = get_uid(object)
        log.debug('adding event source %s'%uid)        
        
        if self._event_sources.has_key(uid):
            if self._event_sources[uid].count(event_type) != 0:
                raise BadEventSource("Source already registered for this event type")
            
            self._event_sources[uid].append(event_type)
        else:
            self._event_sources[uid]=[event_type]
        self._p_changed = 1            

        

    security.declareProtected('Use EventChannel', 'register_source')
    def unregister_source(self, object, event_type=None):
        """ remove a source for a given event channel, if
        no channel given, remove source from all channels """
        
        uid = get_uid(object)
        log.debug('removing event source %s'%uid)
        
        if not self._event_sources.has_key(uid):
            return

        if event_type:
            self._event_sources[uid].remove(event_type)
            if len(self._event_sources[uid]) == 0:
                del self._event_sources[uid]
        else:
            del self._event_sources[uid]
            
        self._p_changed = 1

    security.declarePublic('get_event_sources')
    def get_event_sources(self):
        return self._event_sources.keys()
        
    security.declarePublic('is_source')
    def is_source(self, object):
        return self._event_sources.has_key(get_uid(obj))
    
    security.declareProtected('Manage EventChannel', 'register_event_type') 
    def register_event_type(self, event_type):
        """ """
        if type(event_type) != type(''):
            raise BadEventType( "Event type must be a string")
        elif self._event_types.has_key(event_type):
            raise BadEventType ("%s already exists"%event_type)

        log.debug('adding event type%s'%event_type)        

        self._event_types[event_type] = []
        self._p_changed = 1

    security.declareProtected('Manage EventChannel', 'unregister_event_type') 
    def unregister_event_type(self, event_type):
        """ """
        if type(event_type) != type(''):
            raise BadEventType ("no event type %s"%str(event_type))
        elif not self._event_types.has_key(event_type):
            return
            #raise BadEventType ("no event type %s"%str(event_type))
        
        log.debug("removing event type %s"%event_type)

        # listeners
        listeners = self._event_types.get(event_type, [])
        for l in listeners:
            self.unregister_listener(None, lid=l, event_type=event_type)

        # sources  
        sources = self._event_sources.keys()
        for s in sources:
            if self._event_sources[s].count(event_type) > 0:
                self.unregister_source(None,sid=s, event_type=event_type)

        # event type               
        del self._event_types[event_type]
        
        self._p_changed = 1
        
    security.declarePublic('has_event_type')    
    def has_event_type(self, event_type):
        return self._event_types.has_key(event_type)

    security.declareProtected('Use EventChannel', 'get_event_types')
    def get_event_types(self):
        return self._event_types.keys()

    security.declarePrivate('clear_event_channel')
    def clear_event_channel(self):
        """ clear all """
        self.__init__(self.id, self.title)

    security.declareProtected('Manage EventChannel', 'sync_registrations')
    def sync_registrations(self):
        """ remove unreachable object registrations """
        
        # should also make sure methods and filters are still
        # findable.
        
        sbozos = []
        remove = self._event_sources.__delitem__
        for s in self._event_sources.keys():
            try: sobj = self.unrestrictedTraverse(s)
            except: sbozos.append(s)
        map(remove, sbozos)

        lbozos=[]
        remove = self.unregister_listener
        for l in self._event_listeners.keys():
            try: lobj = self.unrestrictedTraverse(l)
            except: bozos.append(l)
        map(remove, lbozos)
        
        return (len(sbozos), len(lbozos))
        
    security.declareProtected('Publish Event', 'publish')
    def publish(self, obj, event_type):

        log.debug('publishing %s event_type'%event_type)

        
        if not self._event_types.has_key(event_type):
            raise BadEventType (" Event Type not registered ")

        sid = get_uid(obj)
        
        if not self._event_sources.has_key(sid):
            raise BadEventSource (
                " This object is not registered as an event source %s"%sid
            )
        if not self._event_sources[sid].count(event_type) > 0:
            raise BadEventSource (
                "This object is not registered as a source for this event type"
                )
        
        # need per transaction limit controls
        # to prevent mutation chains ie runaways
        self._register()
        
        if hasattr(self.aq_base, '_v_txn_publish_count'):
            if self._v_txn_publish_count > TRANSACTION_PUBLISH_LIMIT:
                raise EventChannelTransactionLimit (
                    """Publishing this event would break per
                    transaction limits %s %s"""%(sid, event_type)
                    )
            else:
                self._v_txn_publish_count += 1
        else:
            self._v_txn_publish_count = 1

            
        listeners = self._event_types[event_type]
        for l in listeners:
            try: lobj = self.unrestrictedTraverse(l)
            except Exception ,e:
                log.critical("obj not found with path %s %s"%(l,str(e)))
                raise e
                #continue # perhaps we should perform stats analysis
                         # here or instead call sync_registrations after..

            log.debug('found object')
            method = self._event_l_methods.get((l,event_type), None)
            if not method:
                log.critical("method not found with for %s %s %s"%(l,event_type, str(e)))
                continue
            else:
                method = getattr(lobj, method, None)
                if not method:
                    raise NoEventTargetFound(""" no event target found for """)

            log.debug('found method')
            
            try: filter = self._get_filter(lobj, l, event_type)
            except Exception, e:            
                log.critical("filter not found with name %s %s %s"%(l, event_type, str(e)))
                continue
                
            log.debug('found filter')
            
            try:
                # filters get listener=self,event obj, event_type1
                if not filter or filter(lobj, obj,event_type) == 1:
                    log.debug('past filter')
                    method(lobj, obj, event_type)
            except Exception, e:
                # hmmm.. programmer error
                log.warning("publish error on %s %s"%(l,str(e)))                
                
        log.debug('done publishing')

    def _get_filter(self, listener_obj, uid, event_type):

        # add some explicit checks for filter discovery?
        
        filter = self._event_filters.get( (uid, event_type), None)
        if filter:
            if filter.count('.') >1:
                idx = filter.rfind('.')                
                func = filter[idx+1:]
                module = filter[:idx]                
                filter = getObject(module, func)
            else:
                filter = getattr(listener_obj, filter)
        
        return filter

    def _validate_filter(self, obj, filter):
        """ Checks to see whether this filter can be found
        with regard to this object, since filters
        can be external methods we check if its
        a path like Bar.Foo or MyProduct.Bar.Foo and then 
        split into (product)module and function and check the 
        extension directories else we check if its a callable 
        attribute that can be found on the object itself 
        (using acquisition).
        """
        if type(obj) == type(''):
            try: obj = self.unrestrictedTraverse(obj)
            except:
                log.critical('unreachable object %s'%obj)
                #raise UnreachableObject()
                return
        \
        if filter:
            
            if type(filter) != type(''):
                raise BadFilterType (
                    " Filters must be strings "
                    )
            if filter.count('.') > 0:
                # lookup ext method basis
                idx = filter.rfind('.')
                func = filter[idx+1:]
                module = filter[:idx]
                # throws errors if not found
                filter = getObject(module, func)
                if not filter:
                    raise NoFilterFound (
                           """ objects must implement registered filters, or
                           filters should be in the context of the object """
                                         )
                    
            elif not hasattr(obj, filter):
                # lookup on attribute basis as string
                
                # should we force filters to only exist
                # upon the object, ie no acq.
                raise NoFilterFound (
                 """ objects must implement registered filters, or
                 filters should be in the context of the object """
                                     )
            
        return 1

        
    # some redirection for handling txn integration
    def _txn_finish(self):
        self._v_eventpublishcount=0
        self._v_txn_registered=0
    def _abort(self):
        self._txn_finish()
    def _finish(self):
        self._txn_finish()
        
        
InitializeClass(EventChannel)




